/* libsswf_file.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/



#include	"sswf/libsswf.h"

using namespace sswf;








/////////////////////////////////////////// Data

Data::Data()
{
	f_pos = 0;
	f_size = 0;
	f_data = 0;
}


Data::~Data()
{
	// the memory manager takes care of everything here
}


void Data::Align(void)
{
	// align to the next byte
	f_pos = (f_pos + CHAR_BIT - 1) & -CHAR_BIT;
}


void Data::SetSize(size_t size)
{
	int		pos;

	if(size > f_size) {
		// got to enlarge the buffer (adjust to the next 256 bytes)
		pos = f_size / CHAR_BIT;
		f_size = (f_pos + size + CHAR_BIT * 256 - 1) & -(CHAR_BIT * 256);
		f_data = (char *) MemRealloc(f_data, f_size / CHAR_BIT, "Data buffer");

		// clear so bits can safely be saved
		memset(f_data + pos, 0, f_size / CHAR_BIT - pos);
	}
}


void Data::Size(unsigned long size)
{
	assert(size <= f_pos, "can't enlarge a data object with a call to the Size() function");

	f_pos = size;
}


void Data::Overwrite(size_t offset, const void *ptr, size_t size)
{
	assert(offset + size < f_pos / CHAR_BIT, "trying to overwrite outside the buffer");
	memcpy(f_data + offset, ptr, size);
}


void Data::OverwriteByte(size_t offset, char c)
{
	assert(offset + 1 < f_pos / CHAR_BIT, "trying to overwrite outside the buffer");
	f_data[offset] = c;
}


void Data::OverwriteShort(size_t offset, short s)
{
	assert(offset + 2 < f_pos / CHAR_BIT, "trying to overwrite outside the buffer");
	f_data[offset] = (char) s;
	f_data[offset + 1] = (char) (s >> 8);
}


void Data::OverwriteLong(size_t offset, long l)
{
	assert(offset + 4 < f_pos / CHAR_BIT, "trying to overwrite outside the buffer");
	f_data[offset    ] = (char) l;
	f_data[offset + 1] = (char) (l >> 8);
	f_data[offset + 2] = (char) (l >> 16);
	f_data[offset + 3] = (char) (l >> 24);
}


void Data::Write(const void *ptr, size_t size)
{
	size_t		bit_size;

	Align();
	bit_size = size * CHAR_BIT;
	SetSize(f_pos + bit_size);
	memcpy(f_data + f_pos / CHAR_BIT, ptr, size);
	f_pos += bit_size;

#if DEBUG
MemTest(f_data);
#endif
}



void Data::WriteBits(long value, size_t bits)
{
	int		top;

	assert(bits == 32 || ((-1 << bits) & value) == 0 || ((-1 << bits) & value) == (-1 << bits),
			"WriteBits() used to write %d bits of 0x%08lX - bits are lost", (int) bits, value);

	SetSize(f_pos + bits);

	top = sizeof(long) * CHAR_BIT - bits;
	if(top != 0) {		// avoid a shift by zero which doesn't always work (system dependent)
		value <<= top;	// put the 1st bit (most significant) in the sign bit
	}

	// TODO: there are much more efficent ways to do that by calculating shift values and
	// shifting 1 to 8 bits in place instead of looping like crazy!!!
	while(bits > 0) {
		bits--;
		if(value < 0) {
			f_data[f_pos / CHAR_BIT] |= (1 << (CHAR_BIT - 1 - (f_pos & (CHAR_BIT - 1))));
		}
		f_pos++;
		value *= 2;
	}
}


void Data::PutByte(char c)
{
	Write(&c, 1);
}


void Data::PutShort(short s)
{
	PutByte((char) s);
	PutByte((char) (s >> 8));
}


void Data::PutLong(long l)
{
	PutByte((char) l);
	PutByte((char) (l >> 8));
	PutByte((char) (l >> 16));
	PutByte((char) (l >> 24));
}


void Data::PutString(const char *string)
{
	if(string == 0) {
		// an empty string
		PutByte(0);
	}
	else {
		Write(string, strlen(string) + 1);
	}
}




void Data::Append(const Data& data)
{
	int		sz;

	Align();
	// data.Align() -- we can't do that on a constant
	// so we compute sz as the aligned f_pos and use
	// that instead
	sz = (data.f_pos + CHAR_BIT - 1) & -CHAR_BIT;
	SetSize(f_pos + sz);
	memcpy(f_data + f_pos / CHAR_BIT, data.f_data, sz / CHAR_BIT);
	f_pos += sz;

#if DEBUG
MemTest(f_data);
#endif
}



void Data::Read(void *& ptr, size_t& size)
{
	Align();
	ptr = f_data;
	size = f_pos / CHAR_BIT;
}





