/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;



/////////////////////////////////////////// Style
Style::Style(void)
{
	Reset();
}


void Style::Reset(void)
{
	int		idx;

	f_style = STYLE_TYPE_UNKNOWN;

	f_morph = false;
	f_use_alpha = false;

	f_line_width[0] = 0;
	f_line_width[1] = 0;

	f_color[0].Reset();
	f_color[1].Reset();

	f_bitmap_ref = 0;

	f_matrix[0].Reset();
	f_matrix[1].Reset();

	f_gradient = -1;	// nothing defined yet
	idx = 16;
	do {
		idx--;
		f_gradient_pos[idx] = 0;
		f_gradient_color[idx].Reset();
	} while(idx > 0);
}



void Style::SetType(style_t style)
{
	switch(f_style) {
	case STYLE_TYPE_LINE:
		assert(style == STYLE_TYPE_LINE, "can't change a line style");
		break;

	case STYLE_TYPE_SOLID:
		assert(style == STYLE_TYPE_SOLID, "can't change a solid fill style");
		break;

	case STYLE_TYPE_GRADIENT_LINEAR:
	case STYLE_TYPE_GRADIENT_RADIAL:
		assert(style == STYLE_TYPE_GRADIENT_LINEAR || style == STYLE_TYPE_GRADIENT_RADIAL,
				"can't change from a gradient fill style to another style");
		break;

	case STYLE_TYPE_BITMAP_TILLED:
	case STYLE_TYPE_BITMAP_CLIPPED:
	case STYLE_TYPE_BITMAP_HARDEDGE_TILLED:
	case STYLE_TYPE_BITMAP_HARDEDGE_CLIPPED:
		assert(style == STYLE_TYPE_BITMAP_TILLED
			|| style == STYLE_TYPE_BITMAP_CLIPPED
			|| style == STYLE_TYPE_BITMAP_HARDEDGE_TILLED
			|| style == STYLE_TYPE_BITMAP_HARDEDGE_CLIPPED,
				"can't change from a bitmap fill style to another style");
		break;

	case STYLE_TYPE_MATRIX:
		assert(style == STYLE_TYPE_GRADIENT_LINEAR
			|| style == STYLE_TYPE_GRADIENT_RADIAL
			|| style == STYLE_TYPE_BITMAP_TILLED
			|| style == STYLE_TYPE_BITMAP_CLIPPED
			|| style == STYLE_TYPE_BITMAP_HARDEDGE_TILLED
			|| style == STYLE_TYPE_BITMAP_HARDEDGE_CLIPPED,
				"can't change from a style with a bitmap or a gradient to one without");
		break;

	case STYLE_TYPE_UNKNOWN:
		break;

	default:
		assert(0, "invalid style");
		/*NOTREACHED*/

	}

	f_style = style;
}


void Style::SetLine(int index, unsigned short width, const Color &color)
{
	assert(index == 0 || index == 1, "invalid index of %d for Style::SetLine()", index);

	SetType(STYLE_TYPE_LINE);

	f_line_width[index] = width;
	f_color[index] = color;

	if(color.Alpha() != 255) {
		f_use_alpha = true;
	}

	if(index == 1) {
		f_morph = true;
	}
}


void Style::SetColor(int index, const Color& color)
{
	assert(index == 0 || index == 1, "invalid index of %d for Style::SetColor()", index);

	SetType(STYLE_TYPE_SOLID);

	f_color[index] = color;

	if(color.Alpha() != 255) {
		f_use_alpha = true;
	}

	if(index == 1) {
		f_morph = true;
	}
}


void Style::SetGradient(int index, int pos, const Color& color)
{
	assert(index >= 0 || index < 16, "invalid index of %d for Style::SetGradient()", index);

	switch(f_style) {
	case STYLE_TYPE_GRADIENT_LINEAR:
	case STYLE_TYPE_GRADIENT_RADIAL:
		break;

	default:
		/* assumes a default */
		SetType(STYLE_TYPE_GRADIENT_LINEAR);
		break;

	}

	f_gradient_pos[index] = pos;
	f_gradient_color[index] = color;

	if(color.Alpha() != 255) {
		f_use_alpha = true;
	}

	if(index >= 8) {
		f_morph = true;
	}

	index &= 7;
	if(index > f_gradient) {
		f_gradient = index;
	}
}




void Style::SetMatrix(int index, const Matrix& matrix)
{
	assert(index == 0 || index == 1, "invalid index of %d for Style::SetMatrix()", index);

	switch(f_style) {
	case STYLE_TYPE_GRADIENT_LINEAR:
	case STYLE_TYPE_GRADIENT_RADIAL:
	case STYLE_TYPE_BITMAP_TILLED:
	case STYLE_TYPE_BITMAP_CLIPPED:
	case STYLE_TYPE_BITMAP_HARDEDGE_TILLED:
	case STYLE_TYPE_BITMAP_HARDEDGE_CLIPPED:
		break;

	default:
		/* set an intermediate value so we can check that people don't change the type on us */
		SetType(STYLE_TYPE_MATRIX);
		break;

	}

	f_matrix[index] = matrix;
}


void Style::SetBitmap(unsigned short id)
{
	assert(id != 0 && id != 65535, "invalid identifier for Style::SetBitmap() - 0 and 65535 are always refused");

	switch(f_style) {
	case STYLE_TYPE_BITMAP_TILLED:
	case STYLE_TYPE_BITMAP_CLIPPED:
	case STYLE_TYPE_BITMAP_HARDEDGE_TILLED:
	case STYLE_TYPE_BITMAP_HARDEDGE_CLIPPED:
		break;

	default:
		/* assumes a default */
		SetType(STYLE_TYPE_BITMAP_TILLED);
		break;

	}

	f_bitmap_ref = id;
}


void Style::SetClipping(void)
{
	switch(f_style) {
	case STYLE_TYPE_BITMAP_TILLED:
	case STYLE_TYPE_BITMAP_CLIPPED:
	case STYLE_TYPE_BITMAP_HARDEDGE_TILLED:
	case STYLE_TYPE_BITMAP_HARDEDGE_CLIPPED:
		break;

	default:
		/* assumes a default */
		SetType(STYLE_TYPE_BITMAP_TILLED);
		break;

	}

	f_bitmap_ref = (sswf_id_t) -1;
}



void Style::Save(Data& data, bool save_alpha, bool save_morph)
{
	int		idx;

	// actually we don't really have to test the following, but for completness it's neat
	assert(save_morph && save_alpha || !save_morph, "can't save morph without alpha");

	if(f_morph ^ save_morph) {
		fprintf(stderr, "WARNING: a style is used with(out) morph information (not) for a DefineMorphShape.\n");
	}
	if(f_use_alpha && !save_alpha) {
		fprintf(stderr, "WARNING: a style using transparency is saved using solid colors.\n");
	}

	switch(f_style) {
	case STYLE_TYPE_LINE:
		data.PutShort(f_line_width[0]);
		if(save_morph) {
			data.PutShort(f_line_width[1]);
			f_color[0].Save(data, true);
			f_color[1].Save(data, true);
		}
		else if(save_alpha) {
			f_color[0].Save(data, true);
		}
		else {
			f_color[0].Save(data, false);
		}
		break;

	case STYLE_TYPE_SOLID:
		data.PutByte(0);		// type SOLID
		if(save_morph) {
			f_color[0].Save(data, true);
			f_color[1].Save(data, true);
		}
		else {
			f_color[0].Save(data, save_alpha);
		}
		break;

	case STYLE_TYPE_GRADIENT_LINEAR:
		data.PutByte(0x10);		// type LINEAR GRADIENT
		goto save_gradient;

	case STYLE_TYPE_GRADIENT_RADIAL:
		data.PutByte(0x12);		// type RADIAL GRADIENT
save_gradient:
		f_matrix[0].Save(data);
		if(save_morph) {
			f_matrix[1].Save(data);
			data.PutByte(f_gradient + 1);
			for(idx = 0; idx <= f_gradient; idx++) {
				data.PutByte(f_gradient_pos[idx]);
				f_gradient_color[idx].Save(data, true);
				data.PutByte(f_gradient_pos[idx + 8]);
				f_gradient_color[idx + 8].Save(data, true);
			}
		}
		else {
			data.PutByte(f_gradient + 1);
			for(idx = 0; idx <= f_gradient; idx++) {
				data.PutByte(f_gradient_pos[idx]);
				f_gradient_color[idx].Save(data, save_alpha);
			}
		}
		break;

	case STYLE_TYPE_BITMAP_TILLED:
		data.PutByte(0x40);
		goto save_bitmap;

	case STYLE_TYPE_BITMAP_CLIPPED:
		data.PutByte(0x41);
		goto save_bitmap;

	case STYLE_TYPE_BITMAP_HARDEDGE_TILLED:
		data.PutByte(0x42);
		goto save_bitmap;

	case STYLE_TYPE_BITMAP_HARDEDGE_CLIPPED:
		data.PutByte(0x43);
save_bitmap:
		data.PutShort(f_bitmap_ref);
		f_matrix[0].Save(data);
		if(save_morph) {
			f_matrix[1].Save(data);
		}
		break;

	default:
		assert(0, "trying to save a style with an undefined/invalid type");
		/*NOTREACHED*/

	}
}



// this is very certainly not necessary at this time
// but it could make things easier in the future...
// (we could compare depending on the style also)
bool Style::operator == (const Style& style) const
{
	int		idx;

	// to simplify the following switch we test this first
	if(f_style != style.f_style || f_morph != style.f_morph) {
		return false;
	}

	// don't test if we compare a style with itself because
	// that's not equivalent to the following!
	switch(f_style) {
	case STYLE_TYPE_LINE:
		return f_line_width[0] == style.f_line_width[0] && f_line_width[1] == style.f_line_width[1]
			&& f_color[0] == style.f_color[0] && f_color[1] == style.f_color[1];

	case STYLE_TYPE_SOLID:
		return f_color[0] == style.f_color[0] && f_color[1] == style.f_color[1];

	case STYLE_TYPE_GRADIENT_LINEAR:
	case STYLE_TYPE_GRADIENT_RADIAL:
		if(f_matrix[0] != style.f_matrix[0] || f_matrix[1] != style.f_matrix[1]
		|| f_gradient != style.f_gradient) {
			return false;
		}
		for(idx = 0; idx < f_gradient; idx++) {
			if(f_gradient_pos[idx    ] != style.f_gradient_pos[idx    ]
			|| f_gradient_pos[idx + 8] != style.f_gradient_pos[idx + 8]
			|| f_gradient_color[idx    ] != style.f_gradient_color[idx    ]
			|| f_gradient_color[idx + 8] != style.f_gradient_color[idx + 8]) {
				return false;
			}
		}
		return true;

	case STYLE_TYPE_BITMAP_TILLED:
	case STYLE_TYPE_BITMAP_CLIPPED:
	case STYLE_TYPE_BITMAP_HARDEDGE_TILLED:
	case STYLE_TYPE_BITMAP_HARDEDGE_CLIPPED:
		return f_matrix[0] == style.f_matrix[0] && f_matrix[1] == style.f_matrix[1]
				&& f_bitmap_ref == style.f_bitmap_ref;

	default:
		assert(0, "trying to compare a style with an undefined type");
		/*NOTREACHED*/

	}

	return true;	// avoid warnings
}




