/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;



/////////////////////////////////////////// TagBase
TagBase::TagBase(const char *name, TagBase *parent)
{
	register TagBase	*tail;

	f_userdata = 0;		// by default we initialize to zero

	f_name = name;
	f_label = 0;		// no label by default, use SetLabel() to set it

	f_parent = parent;
	f_next = 0;
	f_previous = 0;
	f_children = 0;

	f_frames = 0;		// should be reset each time it is to be used!

	// link into the parent if there is one
	if(parent != 0) {
		// the order is VERY important so we HAVE to
		// link at the end of the list
		tail = parent->f_children;
		if(tail == 0) {
			parent->f_children = this;
		}
		else {
			while(tail->f_next != 0) {
				tail = tail->f_next;
			}
			f_previous = tail;
			tail->f_next = this;
		}
	}
}


TagBase::~TagBase()
{
	// if we are being deleted, we ought to delete our children
	// (we may change that at a later time, but it seems as if it were a valid argument right now)
	while(f_children != 0) {
#if DEBUG
#if 0
// IsBuffer() isn't a valid function at this time
fprintf(stderr, "Delete child [%s] [%s] (%d)\n",
				f_children->f_name,
				f_children->f_label == 0 ? "<no label>" : f_children->f_label,
				Buffer::IsBuffer(f_children));
		assert(Buffer::IsBuffer(f_children) == 0, "a tag object can't be attached to another object (name: \"%s\", label: \"%s\")",
				f_children->f_name,
				f_children->f_label == 0 ? "<no label>" : f_children->f_label);
#endif
#endif
		delete f_children;
	}

	if(f_previous == 0) {
		if(f_parent != 0) {
			f_parent->f_children = f_next;
		}
	}
	else {
		f_previous->f_next = f_next;
	}
	if(f_next != 0) {
		f_next->f_previous = f_previous;
	}
}



int TagBase::PreSave(void)
{
	TagBase		*p;
	int		ec;

//printf("--- PreSave %p (%s)\n", this, Name());

	/*
	 * By default we simply call the PreSave() of
	 * every child in this object
	 */
	p = Children();
	while(p != 0) {
		ec = p->PreSave();
		if(ec != 0) {
			return ec;
		}
		p = p->Next();
	}

	return 0;
}



int TagBase::PreSave2ndPass(void)
{
	TagBase		*p;
	int		ec;

//printf("--- PreSave2ndPass %p (%s)\n", this, Name());

	/*
	 * By default we simply call the PreSave2ndPass() of
	 * every child in this object
	 */
	p = Children();
	while(p != 0) {
		ec = p->PreSave2ndPass();
		if(ec != 0) {
			return ec;
		}
		p = p->Next();
	}

	return 0;
}


int TagBase::SaveString(Data& data, const char *string)
{
	TagHeader *header = Header();
	if(header != 0) {
		return header->SaveEncodedString(data, string);
	}
	return 1;
}



sswf_frame_t TagBase::WhichFrame(void) const
{
	sswf_frame_t	result;
	const TagBase	*p;

	if(strcmp(f_name, "header") == 0 || f_parent == 0) {
		/* there can't really be a frame in a header tag!!! */
		return 0;
	}

	if(strcmp(f_parent->f_name, "header") != 0 && strcmp(f_parent->f_name, "sprite") != 0) {
		/* at this time other possible cases aren't supported... */
		return 0;
	}

	result = 0;
	p = this->f_previous;	/* start with the previous since we don't want to include the current ShowFrame */
	while(p != 0) {
		if(strcmp(p->f_name, "showframe") == 0) {
			result++;
		}
		p = p->f_previous;
	}

	return result;
}


void TagBase::SetLabel(const char *label)
{
	MemFree(f_label);
	f_label = StrDup(label);
}


TagBase *TagBase::FindLabel(const TagBase *p, const char *label) const
{
	const TagBase		*n;

	while(p->f_previous != 0) {	// should always be false!
		p = p->f_previous;
	}
	// search this level first
	while(p != 0) {
		if(p->f_label != 0 && strcmp(p->f_label, label) == 0) {
			return const_cast<TagBase *>(p);
		}
		if(p->f_children != 0) {
			n = FindLabel(p->f_children, label);
			if(n != 0) {
				return const_cast<TagBase *>(n);
			}
		}
		p = p->f_next;
	}

	return 0;
}


TagBase *TagBase::FindLabelledTag(const char *label) const
{
	// this search looks in the entire tree available
	const TagBase		*p;

	// start searching from the root tag
	p = this;
	if(p == 0) {
		return 0;		// works when this is still a nul pointer!
	}
	while(p->f_parent != 0) {
		p = p->f_parent;
	}

	if(label == 0 || label[0] == '\0') {
		return const_cast<TagBase *>(p);		// return the root in this case
	}

	return FindLabel(p, label);
}


TagBase *TagBase::FindID(const TagBase *p, sswf_id_t id, bool search_import) const
{
	const TagBase		*n;

	while(p->f_previous != 0) {	// should always be false!
		p = p->f_previous;
	}
	// search this level first
	while(p != 0) {
		if((p->TypeFlags() & SWF_TYPE_HAS_ID) != 0) {
			if(dynamic_cast<const TagBaseID *>(p)->Identification() == id) {
				return const_cast<TagBase *>(p);
			}
		}
		else if(search_import && strcmp(p->Name(), "import") == 0) {
			// import objects can offer multiple IDs
			if(dynamic_cast<const TagImport *>(p)->HasID(id) != 0) {
				return const_cast<TagBase *>(p);
			}
		}
		if(p->f_children != 0) {
			n = FindID(p->f_children, id, search_import);
			if(n != 0) {
				return const_cast<TagBase *>(n);
			}
		}
		p = p->f_next;
	}

	return 0;
}


TagBase *TagBase::FindTagWithID(sswf_id_t id, bool search_import) const
{
	// this search looks in the entire tree available
	const TagBase		*p;

	// start searching from the root tag
	p = this;
	if(p == 0) {
		return 0;		// works when this is still a nul pointer!
	}
	while(p->f_parent != 0) {
		p = p->f_parent;
	}

	return FindID(p, id, search_import);
}


int TagBase::SaveTag(Data& data, swf_tag_t tag, size_t size)
{
	if(size >= 63) {		// large tag?
		data.PutShort((tag << 6) + 63);
		data.PutLong(size);
	}
	else {
		data.PutShort((tag << 6) + size);
	}

	return 0;
}




long TagBase::SIBitSize(long value)
{
	long	cnt;

#if 0
	if(value < 0) {
		// a special case
		if(value == -1) {
			return 1;
		}
		cnt = 33;
		do {
			cnt--;
			value *= 2;
		} while(value < 0);
	}
	else {
		// a special case
		if(value == 0) {
			return 1;
		}
		cnt = 33;
		do {
			cnt--;
			value *= 2;
		} while(value > 0);
	}
#endif
	// fix by Adam Polkosnik who has a 64 bit processor
	if(value < 0) {
		value = ~value;
	}

	cnt = 1;
	while(value > 0) {
		cnt++;
		value /= 2;
	}

	return cnt;
}


long TagBase::UIBitSize(unsigned long value)
{
	long	cnt;

#if 0
	// a special case
	if(value == 0) {
		return 1;
	}
	cnt = 32;
	while((long) value > 0) {
		cnt--;
		value *= 2;
	}
#endif

	// fix by Adam Polkosnik who has a 64 bit processor
	cnt = 0;
	do {
		cnt++;
		value /= 2;
	} while(value > 0);

	return cnt;
}


long TagBase::Double2Signed(double value)
{
	// assert(?, "overflow"); -- should we otherwise have a bound checking mechanism
	return (long) rint(value * 65536.0);
}


double TagBase::Signed2Double(long value)
{
	return (double) value / 65536.0;
}


void TagBase::ShowFrame(void)
{
	f_frames++;
}



void TagBase::MinimumVersion(unsigned char version)
{
	Header()->MinimumVersion(version);
}


unsigned char TagBase::Version(void) const
{
	return Header()->Version();
}


/* this is a pure virtual function
TagBase::swf_type_t TagBase::TypeFlags(void) const
{
	fprintf(stderr, "TypeFlag() undefined for object type \"%s\".\n", Name());
	abort();
}
*/


TagHeader *TagBase::Header(void) const
{
	const TagBase		*p;

	// when in the header, we return the header!
	p = this;
	while(p != 0) {
		if(strcmp(p->Name(), "header") == 0) {
			return const_cast<TagHeader *>(dynamic_cast<const TagHeader *>(p));
		}
		p = p->f_parent;
	}

	return 0;
}


// The name of a tag is actually its type (header, text, edit, doaction, ...)
const char *TagBase::Name(void) const
{
	return f_name;
}


const char *TagBase::Label(void) const
{
	return f_label;
}


void TagBase::ResetFrames(void)
{
	f_frames = 0;
}


sswf_frame_t TagBase::FrameCount(void) const
{
	return f_frames;
}

TagBase *TagBase::Children(void)
{
	return f_children;
}


TagBase *TagBase::Next(void)
{
	return f_next;
}


TagBase *TagBase::Previous(void)
{
	return f_previous;
}


TagBase *TagBase::Parent(void)
{
	return f_parent;
}



/////////////////////////////////////////// TagBaseID
TagBaseID::TagBaseID(const char *name, TagBase *parent)
	: TagBase(name, parent)
{
	assert(parent != 0, "a tag with an identification must have a header as its parent");

	f_id = Header()->NextID();
	f_identified = true;
}


sswf_id_t TagBaseID::Identification(void) const
{
	return f_id;
}


void TagBaseID::NoIdentification(void)
{
	if(f_identified) {
		Header()->RemoveID(f_id);
		f_identified = false;
	}
}


int TagBaseID::SaveID(Data& data) const
{
	assert(f_identified, "the identification of this object was removed");
	data.PutShort(f_id);

	return 0;
}


