/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;



/////////////////////////////////////////// State
State::State(void)
{
	Reset();
}


void State::Reset(void)
{
	f_id = 0;
	f_flags = 0;		// invalid value!
	f_layer = 0;
	f_matrix.Reset();
	f_color_transform.Reset();
	f_has_color_transform = false;
}

bool State::HasColorTransform(void) const
{
	return f_has_color_transform;
}


void State::SetObjectID(sswf_id_t id)
{
	f_id = id;
	f_id_defined = true;
}


void State::SetFlags(unsigned char flags)
{
	f_flags = flags;
}


void State::SetLayer(unsigned short layer)
{
	f_layer = layer;
}


void State::SetMatrix(const Matrix& matrix)
{
	f_matrix = matrix;
}


void State::SetColorTransform(const ColorTransform& color_transform)
{
	f_color_transform = color_transform;
	f_has_color_transform = true;
}


void State::Save(Data& data, bool color_transform)
{
	assert(f_flags != 0, "invalid flags for a State object to be saved");

	data.PutByte(f_flags);		// these are the 8 first bits
	data.PutShort(f_id);
	data.PutShort(f_layer);
	f_matrix.Save(data);

	if(color_transform) {
		f_color_transform.Save(data);
	}
}





/////////////////////////////////////////// TagButton
TagButton::TagButton(TagBase *parent)
	: TagBaseID("button", parent)
{
}


TagBase::swf_type_t TagButton::TypeFlags(void) const
{
	return SWF_TYPE_DEFINE | SWF_TYPE_REFERENCE | SWF_TYPE_SCRIPT | SWF_TYPE_HAS_ID;
}


void TagButton::SetState(const State& state)
{
	State		*s;

	s = new State(state);
	MemAttach(s, sizeof(State), "TagButton::SetState() -- state duplicate in Button tag");

	f_states.Set(-1, s);
}


void TagButton::SetAction(const Action& action)
{
	Action		*a;

	a = action.Duplicate();
	MemAttach(a, sizeof(Action), "TagButton::SetAction() -- action duplicate in Button tag");

	f_actions.Set(-1, a);
}


Vectors& TagButton::Actions(void)
{
	return f_actions;
}


void TagButton::AddEvent(Event *event)
{
	f_events.Set(-1, event);
}


void TagButton::SetMenu(bool menu)
{
	f_menu = menu != 0;
}


int TagButton::PreSave(void)
{
	int		version, v, max;
	State		*s;
	Event		*e;

// by default we save as a DefineButton (v1.x)
	f_save_button2 = false;
	version = 1;

// check all the actions
	v = Action::MinimumListVersion(f_actions);
	if(v > version) {
		version = v;
	}

// when there are events, we must check each of their actions
	max = f_events.Count();
	if(max > 0) {
		f_save_button2 = true;
		while(max > 0) {
			max--;
			e = dynamic_cast<Event *>(f_events.Get(max));
			v = Action::MinimumListVersion(e->Actions());
			if(v > version) {
				version = v;
			}
		}
	}

// the f_menu flag means we want at least v3.x
	if(f_menu) {
		f_save_button2 = true;
	}

// if one or more state has a color transform we need a v3.x
	max = f_states.Count();
	if(max == 0) {
		// TODO: we need to have an error manager object!
		fprintf(stderr, "ERROR: can't save a Button tag without having at least one state.\n");
		return -1;
	}
	while(max > 0) {
		max--;
		s = dynamic_cast<State *>(f_states.Get(max));
		if(s->HasColorTransform()) {
			// if we find one, that's it we're done here
			f_save_button2 = true;
			break;
		}
	}

	if(f_save_button2 && version < 3) {
		version = 3;
	}

	MinimumVersion(version);

	return 0;
}


int TagButton::Save(Data& data)
{
	Data		sub_data, extra_data;
	int		idx, max, ec, release;
	unsigned long	cond;
	State		*s;
	Event		*e;

// this object ID
	SaveID(sub_data);

	if(f_save_button2) {
		// at this time there is only one bit (v6.x),
		// in a later version we may need to change the following line.
		sub_data.PutByte(f_menu);
	}

// save the states
	max = f_states.Count();
	if(max == 0) {
		fprintf(stderr, "ERROR: can't save a Button tag without having at least one state.\n");
		return -1;
	}
	for(idx = 0; idx < max; idx++) {
		s = dynamic_cast<State *>(f_states.Get(idx));
		s->Save(extra_data, f_save_button2);
	}
	extra_data.PutByte(0);	/* terminates the list of button states */

// in DefineButton2 we have a list of conditions + actions
// instead of just a list of actions
	if(f_save_button2) {
// in DefineButton2 we save the size of the states before the states themselves
// (so the player can skip them without having to parse them first)
		sub_data.PutShort((short) (extra_data.ByteSize() + 2));
		sub_data.Append(extra_data);

		// handle the special case where a person uses
		// actions as is in the button by inserting these
		// as a POINTER_RELEASE event
		release = 0;
		max = f_events.Count();
		for(idx = 0; idx < max; idx++) {
			extra_data.Empty();
			e = dynamic_cast<Event *>(f_events.Get(idx));
			cond = e->Conditions();
			extra_data.PutShort((short) cond);
			if((cond & Event::CONDITION_POINTER_RELEASE_INSIDE) != 0 && f_actions.Count() > 0) {
				ec = Action::SaveList(Parent(), &f_actions, extra_data, &e->Actions());
				release = 1;
			}
			else {
				ec = Action::SaveList(Parent(), &e->Actions(), extra_data);
			}
			if(ec != 0) {
				return ec;
			}
			if(idx + 1 == max && (f_actions.Count() == 0 || release == 1)) {
				// last size is always zero
				sub_data.PutShort(0);
			}
			else {
				sub_data.PutShort((short) (extra_data.ByteSize() + 2));
			}
			sub_data.Append(extra_data);
		}
		if(f_actions.Count() > 0 && release == 0) {
			extra_data.Empty();
			sub_data.PutShort(0);		// last size is always zero
			sub_data.PutShort(Event::CONDITION_POINTER_RELEASE_INSIDE);
			ec = Action::SaveList(Parent(), &f_actions, sub_data);
			if(ec != 0) {
				return ec;
			}
		}
		//sub_data.PutShort(0);	/* terminates the list of button conditions */
	}
	else {
// in DefineButton we save the states as is
		sub_data.Append(extra_data);

// save the actions
		ec = Action::SaveList(Parent(), &f_actions, sub_data);
		if(ec != 0) {
			return ec;
		}
	}

// create the tag
	SaveTag(data, f_save_button2 ? SWF_TAG_DEFINE_BUTTON2 : SWF_TAG_DEFINE_BUTTON, sub_data.ByteSize());
	data.Append(sub_data);

	return 0;
}




