/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;



/////////////////////////////////////////// TagHeader

TagHeader::TagHeader(void)
	: TagBase("header", 0)
{
	f_version = 0;
	f_min_version = 0;
	f_frame.Set(0, 100 * 20, 0, 100 * 20);		// width  height = 100  100 (in pixels)
	f_rate = 30.0;
	f_compress = false;
	f_next_id = 0;
	f_output_encoding = 0;
	f_iconvertor_open = false;
	f_iconvertor = (iconv_t) -1;	// if f_iconvertor_open is false then this won't be used

#if DEBUG
	// this shouldn't be used until the Save() function is called
	f_saving = false;
#endif

}

TagHeader::~TagHeader()
{
	if(f_iconvertor_open) {
		iconv_close(f_iconvertor);
	}
}



unsigned short TagHeader::NextID(void)
{
	f_next_id++;
	return f_next_id;
}


void TagHeader::RemoveID(sswf_id_t id)
{
	// we restore this ID only if possible, otherwise forget
	// about it altogether (we could create a list of "removed"
	// IDs so we could reuse them later... anyway!)
	if(f_next_id == id) {
		f_next_id--;
	}
}


void TagHeader::SetFrame(const SRectangle& rect)
{
	// the frame rectangle must have the minimum X and Y positions
	// set to zero so we use the total width and height of the
	// user rectangle instead of the real positions
	f_frame.SetReorder(0, labs(rect.XMin() - rect.XMax()), 0, labs(rect.YMin() - rect.YMax()));
}


const SRectangle& TagHeader::Frame(void) const
{
	return f_frame;
}


void TagHeader::SetOutputEncoding(const char *encoding)
{
	MemFree(f_output_encoding);
	if(encoding == 0 || *encoding == '\0') {
		f_output_encoding = 0;
	}
	else {
		f_output_encoding = StrDup(encoding);
	}
}


void TagHeader::SetRate(float rate)
{
	f_rate = rate;
}


float TagHeader::Rate(void) const
{
	return f_rate;
}


void TagHeader::SetAutoOrder(bool auto_order)
{
	f_auto_order = auto_order;
}


void TagHeader::SetCompress(bool compress)
{
	f_compress = compress;
}


int TagHeader::DefineMinimumVersion(void)
{
	unsigned char	v;
	int		ec;

#if DEBUG
	// in pre-save mode we can change the minimum version
	f_saving = false;
#endif

	// the minimum version is dynamically determined
	// we start with a version of 1
	// (this should change to the minimum version that
	// the user wants)
	if(f_version == 0) {
		f_min_version = f_compress ? 6 : 1;
	}
	else {
		if(f_compress && f_version < 6) {
			// can't save a compressed movie in a
			// version earlier than v6.x
			return -1;
		}
		f_min_version = f_version;
	}

	// ensure everything is properly setup in all the
	// object by a call to the PreSave()
	ec = PreSave();
	if(ec != 0) {
		return -labs(ec);
	}

	v = f_min_version;
	ec = PreSave2ndPass();
	if(ec != 0) {
		return -labs(ec);
	}

	// we can't safely switch between any versions in
	// the 2nd pass -- the fonts wouldn't be correct
	// otherwise, so we have to call the pre-save functions
	// all over again!
	if(v <= 5 && f_min_version >= 6) {
		// this time we start with f_min_version >= 6!
		ec = PreSave();
		if(ec != 0) {
			return -labs(ec);
		}

		ec = PreSave2ndPass();
		if(ec != 0) {
			return -labs(ec);
		}
	}

#if DEBUG
	// in save mode we cannot change the minimum version
	f_saving = true;
#endif

	return (int) f_min_version;
}


int TagHeader::Save(Data& data)
{
	TagBase		*child, *last;
	Data		result, header;
	int		ec;
	size_t		length, children_size;
	uLongf		l;
	Bytef		*d;
	void		*ptr;

	// defining the minimum version calls the PreSave[2ndPass]() functions
	// as required before a save
	ec = DefineMinimumVersion();
	if(ec < 0) {
		return ec;
	}

	// reset the counter in case the user saves multiple times
	ResetFrames();

	// save all the children first
	last = 0;
	child = Children();
	while(child != 0) {
		ec = child->Save(result);
		if(ec != 0) {
			return ec;
		}
		last = child;
		child = child->Next();
	}

	if(f_min_version > f_version && f_version != 0) {
		return 1;
	}

	if(last == 0) {
		// there is nothing in there?!!?
		return 1;
	}

	// add the SHOW FRAME & END tags so it works properly with any player
	// NOTE: if the list ends with "end" and that "end" isn't preceeded
	//       by show_frame, then there won't be a show frame.
	if(strcmp(last->Name(), "end") != 0) {
		if(strcmp(last->Name(), "showframe") != 0) {
			ShowFrame();
			SaveTag(result, SWF_TAG_SHOW_FRAME, 0);
		}
		result.PutShort(SWF_TAG_END);
	}
	result.Align();

	// in result we have got all the info now
	// create the header itself

	// save the movie sizes
	f_frame.Save(header);
	header.Align();
	header.PutShort((unsigned short) floor(f_rate * 256.0));
	header.PutShort(FrameCount());

	children_size = result.ByteSize();
	length = children_size + header.ByteSize() + 8;	/* 8 bytes is the size of the header without the rectangle which is already included in the header */

	data.PutByte(f_compress ? 'C' : 'F');
	data.PutByte('W');
	data.PutByte('S');
	data.PutByte(f_version == 0 ? f_min_version : f_version);
	data.PutLong(length);

	if(f_compress) {
		// v6.x+ only
		// in this case we need to merge the header + result
		// and compress it before to append to the data buffer
		header.Append(result);		// TODO: this is a bit ulgy I think...

		header.Read(ptr, length);
		l = (length * 11) / 10 + 256;		// buffer + 10% + 256 bytes
		d = (Bytef *) MemAlloc(l, "compressed movie buffer");
		compress(d, &l, (const Bytef *) ptr, length);
		data.Write(d, l);
		MemFree(d);
	}
	else {
		// not compressed, append everything as is
		data.Append(header);
		data.Append(result);
	}

	return 0;
}



int TagHeader::SaveEncodedString(Data& data, const char *string)
{
	char		buf[256], *s, *p;
	const char	*encoding;
	size_t		out, in;
	int		r;

#if DEBUG
	// here we must be in the save process
	assert(f_saving, "can't call SaveEncodedString(data, \"%s\") without being within a Save() function", string);
#endif

	if(string == 0 || *string == '\0') {
		// nothing to encode in an empty string
		data.PutByte(0);
		return 0;
	}

	if((f_version == 0 ? f_min_version : f_version) >= 6) {
		// version 6.x takes UTF-8 strings which is exactly what
		// we have by default
		data.PutString(string);
		return strlen(string);
	}

	// we need the iconv() call to encode the string as the
	// user specified before to save it in the Data() buffer
	if(!f_iconvertor_open) {
		// convertor not opened yet!
		encoding = f_output_encoding == 0 ? "iso_8859-1" : f_output_encoding;
		f_iconvertor = iconv_open(encoding, "UTF-8");
		if(f_iconvertor == (iconv_t) -1) {
			fprintf(stderr, "ERROR: can't open encoder to convert characters from \"UTF-8\" to \"%s\".\n", encoding);
			return -1;
		}
		f_iconvertor_open = true;
	}
	in = strlen(string);
	out = in * 16;
	if(out >= sizeof(buf)) {
		// use an intermediate string we have to allocate
		s = (char *) MemAlloc(out, "SaveEncodedString(): intermediate string buffer");
	}
	else {
		s = buf;
	}
	p = s;

#if 0
{
printf("Before:");
const unsigned char *t = (const unsigned char *) string;
while(*t != '\0') {
	printf(" %02X", *t);
	t++;
}
}
#endif

	r = (int) iconv(f_iconvertor, ICONV_INPUT_CAST &string, &in, &p, &out);
	if(r < 0) {
		if(s != buf) {
			MemFree(s);
		}
		fprintf(stderr, "ERROR: can't convert string \"%s\".\n", string);
		return -1;
	}
	*p = '\0';

#if 0
{
printf(", after:");
const unsigned char *t = (const unsigned char *) s;
while(*t != '\0') {
	printf(" %02X", *t);
	t++;
}
}
#endif

	data.PutString(s);
	r = strlen(s);
	if(s != buf) {
		MemFree(s);
	}

	return r;
}




void TagHeader::MinimumVersion(unsigned char version)
{
	if(this == 0) {
		return;
	}

#if DEBUG
	// this is to make sure we don't call this function while
	// saving since changing the version may actually invalidate
	// the previous tags in the movie
	assert(!f_saving, "TagHeader::MinimumVersion() called while saving movie -- this is illegal");
#endif

	if(version > f_min_version) {
		f_min_version = version;
	}
}


unsigned char TagHeader::Version(void) const
{
	return f_version == 0 ? f_min_version : f_version;
}



void TagHeader::SetVersion(unsigned char version)
{
	f_version = version;
}


TagBase::swf_type_t TagHeader::TypeFlags(void) const
{
	return SWF_TYPE_HEADER | SWF_TYPE_UNIQUE | SWF_TYPE_START;
}


