/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;




/////////////////////////////////////////// TagShape
TagShape::TagShape(TagBase *parent)
	: TagBaseID("shape", parent)
{
	f_version = 1;
	f_morph = false;
	f_is_glyph = false;
	f_show_bounds = false;
	f_show_origin = false;
	// f_bounds[0/1] -- auto-init
	// f_bounds
	// f_shapes -- auto-init
	f_edges = 0;
	f_morph_edges = 0;
	f_setup = 0;
	// f_fill_styles -- auto-init
	// f_line_styles -- auto-init
	// f_record -- auto-init
	// f_morph_record -- auto-init
}


TagBase::swf_type_t TagShape::TypeFlags(void) const
{
	return SWF_TYPE_DEFINE | SWF_TYPE_REFERENCE | SWF_TYPE_HAS_ID;
}


void TagShape::RecordEdges(void)
{
	if(f_edges != 0) {
		f_record.Set(-1, f_edges);
		f_edges = 0;
	}
}



void TagShape::RecordSetup(void)
{
	if(f_setup != 0) {
		f_record.Set(-1, f_setup);
		f_setup = 0;
	}
}



void TagShape::NewEdges(void)
{
	RecordSetup();

	if(f_edges == 0) {
		f_edges = new shape_edges_t(SHAPE_EDGES);
		MemAttach(f_edges, sizeof(shape_edges_t), "TagShape::NewEdges() -- shape edges array");
		//f_edges->f_edges -- initialized by the new operator
	}
}


void TagShape::NewSetup(void)
{
	RecordEdges();

	if(f_setup == 0) {
		f_setup = new shape_setup_t(SHAPE_SETUP);
		MemAttach(f_setup, sizeof(shape_setup_t), "TagShape::NewSetup() -- shape setup info");
	}
}



void TagShape::AddEdge(morph_t morph_mode, const Edges::edge_t& edge)
{
	assert(morph_mode >= 0 && morph_mode <= 2, "the morph_mode parameter can only be 0, 1 or 2");

	if(morph_mode == MORPH_MODE_SHAPE1 || morph_mode == MORPH_MODE_BOTH_SHAPES) {
		SetMorph();

		if(f_morph_edges == 0) {
			f_morph_edges = new shape_edges_t(SHAPE_EDGES);
			MemAttach(f_morph_edges, sizeof(shape_edges_t), "TagShape::AddEdge() -- shape morph edges array");
			//f_morph_edges->f_edges -- initialized by the new operator
		}

		f_morph_edges->f_edges.Set(edge);
	}

	if(morph_mode == MORPH_MODE_SHAPE0 || morph_mode == MORPH_MODE_BOTH_SHAPES) {
		// we need at least one style
		assert(f_fill_styles.Count() > 0 || f_line_styles.Count() > 0, "can't insert an edge without any style");

		NewEdges();
		f_edges->f_edges.Set(edge);
	}
}


void TagShape::AddStyle(const Style& style, int fill)
{
	// when a new style is given, we first search in the existing styles to see if
	// it is the first time it is inserted; if not, then we use the previous one
	// (we avoid duplicating the same style(s) many times!)
	int		idx;
	const Style	*s;
	Style		*keep;

	NewSetup();

	if(style.HasAlpha() && f_version < 3) {
		f_version = 3;
	}
	if(style.HasHardEdges() && f_version < 7) {
		f_version = 7;
	}
	if(style.HasMorph()) {
		SetMorph();
	}

	idx = 0;
	switch(style.Type()) {
	case Style::STYLE_TYPE_NO_LINE:
		f_setup->f_line_ref = 0;
		break;

	case Style::STYLE_TYPE_LINE:
		idx = f_line_styles.Count();
		while(idx > 0) {
			idx--;
			s = dynamic_cast<const Style *>(f_line_styles.Get(idx));
			if(style == *s) {
				// use that already saved style
				goto add_line;
			}
		}
		// it doesn't exist, we need to insert it now
		idx = f_line_styles.Count();
		// TODO: if the Count() is 32768 then we need to reset the counters...
		//	 don't forget to copy the current fill styles
		//	 See the NewStyles() function
		// NOTE: if the maximum version is V1.0, then the limit should be
		//	 set to 255 instead
		assert(idx < 32766, "too many line styles");
		keep = new Style;
		MemAttach(keep, sizeof(Style), "TagShape::AddStyle() -- create a style (line)");
		*keep = style;			// copy the style
		f_line_styles.Set(-1, keep);
add_line:
		f_setup->f_line_ref = idx + 1;
		break;

	case Style::STYLE_TYPE_NO_FILL:
		assert(fill == 0 || fill == 1, "TagShape::AddStyle() -- invalid fill parameter");
		f_setup->f_fill_ref[fill] = 0;
		break;

	default:	// most other cases...
		assert(fill == 0 || fill == 1, "TagShape::AddStyle() -- invalid fill parameter");
		idx = f_fill_styles.Count();
		while(idx > 0) {
			idx--;
			s = dynamic_cast<const Style *>(f_fill_styles.Get(idx));
			if(style == *s) {
				// use that already saved style
				goto add_fill;
			}
		}
		// it doesn't exist, we need to insert it now
		idx = f_fill_styles.Count();
		// TODO: if the Count() is 32768 then we need to reset the counters...
		//	 don't forget to copy the other current fill & line styles
		//	 See the NewStyles() function
		// NOTE: if the maximum version is V1.0, then the limit should be
		//	 set to 255 instead
		assert(idx < 32766, "too many fill styles");
		keep = new Style;
		MemAttach(keep, sizeof(Style), "TagShape::AddStyle() -- create a style (fill)");
		*keep = style;			// copy the style
		f_fill_styles.Set(-1, keep);
add_fill:
		f_setup->f_fill_ref[fill] = idx + 1;
		break;

	}

	// overflow for V1.0 of Flash
	if(idx > 255 && f_version < 2) {
		f_version = 2;
	}
}


void TagShape::AddMove(morph_t morph_mode, long x, long y)
{
	if(morph_mode == MORPH_MODE_SHAPE1 || morph_mode == MORPH_MODE_BOTH_SHAPES) {
		SetMorph();

		if(f_morph_edges != 0) {
			f_morph_record.Set(-1, f_morph_edges);
			f_morph_edges = 0;
		}

		shape_setup_t *setup = new shape_setup_t(SHAPE_SETUP);
		MemAttach(setup, sizeof(shape_setup_t), "TagShape::AddMove() -- shape morph setup info (i.e. move only)");
		setup->f_x = x;
		setup->f_y = y;
		f_morph_record.Set(-1, setup);
	}

	if(morph_mode == MORPH_MODE_SHAPE0 || morph_mode == MORPH_MODE_BOTH_SHAPES) {
		NewSetup();

		f_setup->f_x = x;
		f_setup->f_y = y;
	}
}


void TagShape::SetMorph(void)
{
	f_morph = true;
	if(f_version < 3) {
		f_version = 3;
	}
}


void TagShape::SetBounds(int index, const SRectangle& rect, bool show_bounds)
{
	assert(index == 0 || index == 1, "invalid index for TagShape::SetBounds()");

	f_show_bounds = show_bounds;
	f_bounds[index] = rect;
	if(index == 1) {
		SetMorph();
	}
}


void TagShape::NewStyles(void)
{
	// this will enable one to have smaller fill & line style
	// references (less bits) -- however that's most probably
	// useless (my point of view 8-)
	shape_record_t	*sr;

	// valid only in V2.0+
	if(f_version < 2) {
		f_version = 2;
	}

	sr = new shape_record_t;
	MemAttach(sr, sizeof(shape_record_t), "TagShape::NewStyles() -- shape record used to have new styles");

	sr->f_fill_styles = new Vectors(f_fill_styles);
	MemAttach(sr->f_fill_styles, sizeof(Vectors), "TagShape::NewStyles() -- fill styles record copy");
	f_fill_styles.Empty();

	sr->f_line_styles = new Vectors(f_line_styles);
	MemAttach(sr->f_line_styles, sizeof(Vectors), "TagShape::NewStyles() -- line styles record copy");
	f_line_styles.Empty();

	sr->f_record = new Vectors(f_record);
	MemAttach(sr->f_record, sizeof(Vectors), "TagShape::NewStyles() -- shape record copy");
	f_record.Empty();

	f_shapes.Set(-1, sr);
}



void TagShape::Glyph(void)
{
	f_is_glyph = true;
	NoIdentification();
}


bool TagShape::IsGlyph(void) const
{
	return f_is_glyph;
}


bool TagShape::IsEmpty(void) const
{
	return f_record.Count() == 0 && f_edges == 0;
}


int TagShape::Save(Data& data)
{
	// In order to save a shape properly we need to (1) know its
	// version (f_version) and whether it is a morphing shape
	// (f_morph).
	save_info_t	info, sub_info, *info_ptr;
	swf_tag_t	tag;
	int		ec, idx, max;
	shape_setup_t	last_setup(SHAPE_SETUP, true);
	SRectangle	rect;

	RecordEdges();
	// RecordSetup(); -- don't record the last setup!!! that's totally useless!!!

	if(f_is_glyph) {
		// save nothing for a glyph since these are saved within a font instead
		return 0;
	}

	// default for V1.0
	tag = SWF_TAG_DEFINE_SHAPE;
	info.f_save_alpha = false;
	info.f_ext_size = false;

	switch(f_version) {
	case 2:
		//info.f_save_alpha = false;
		info.f_ext_size = true;
		tag = SWF_TAG_DEFINE_SHAPE2;
		break;

	case 3:
	case 7:
		info.f_save_alpha = true;
		info.f_ext_size = true;
		if(f_morph) {
			tag = SWF_TAG_DEFINE_MORPH_SHAPE;
		}
		else {
			tag = SWF_TAG_DEFINE_SHAPE3;
		}
		break;

#if DEBUG
	case 1:
		break;

	default:
		assert(0, "invalid f_version in a TagShape object to be saved");
		/*NOTREACHED*/
#endif

	}
	info.f_first = true;
	info.f_save_styles = true;

	if(f_morph && f_bounds[1].IsEmpty()) {
		f_bounds[1] = f_bounds[0];
	}

	SaveID(info.f_data);
	max = f_morph ? 2 : 1;
	for(idx = 0; idx < max; idx++) {
		info.f_data.Align();
		if(f_show_bounds) {
			// first we have to make sure that it is properly ordered
			rect.SetReorder(f_bounds[idx].XMin(), f_bounds[idx].XMax(), f_bounds[idx].YMin(), f_bounds[idx].YMax());
			// then we can enlarge it as required to include the bounding line
			rect.Set(rect.XMin() - 2, rect.XMax() + 2, rect.YMin() - 2, rect.YMax() + 2);
			rect.Save(info.f_data);
		}
		else {
			f_bounds[idx].Save(info.f_data);
		}
	}

	if(f_morph) {
		// in case of a morph we need to save all the styles
		// and the first shape to know the size to save in
		// the offset;
		sub_info.f_save_alpha = info.f_save_alpha;
		sub_info.f_ext_size = info.f_ext_size;
		sub_info.f_first = true;
		sub_info.f_save_styles = true;

		info_ptr = &sub_info;
	}
	else {
		info_ptr = &info;
	}

	// save the extras
	max = f_shapes.Count();
	for(idx = 0; idx < max; idx++) {
		info_ptr->f_sr = *dynamic_cast<shape_record_t *>(f_shapes.Get(idx));
		ec = SaveShape(*info_ptr, last_setup);
		if(ec != 0) {
			return ec;
		}
	}

	// now save the current
	info_ptr->f_sr.f_fill_styles = &f_fill_styles;
	info_ptr->f_sr.f_line_styles = &f_line_styles;
	info_ptr->f_sr.f_record = &f_record;
	ec = SaveShape(*info_ptr, last_setup);
	if(ec != 0) {
		return ec;
	}


	if(f_morph) {
		// the last setup for the morphed version has
		// to start from scratch
		shape_setup_t	last_setup(SHAPE_SETUP, true);

		// save an end of record now (between both shapes)
		sub_info.f_data.WriteBits(0, 6);

		// save the offset to the 2nd set of edges
		info.f_data.PutLong(sub_info.f_data.ByteSize());

		// now we can add the style & 1st set of edges to the main buffer
		info.f_data.Append(sub_info.f_data);

		// we need to save a count for the styles; since they are
		// not referenced at all, we always use 0x11
		info.f_data.PutByte(0x11);

		// we need to record the last edges now
		if(f_morph_edges != 0) {
			f_morph_record.Set(-1, f_morph_edges);
			f_morph_edges = 0;
		}

		// now save the morphed edges
		info.f_sr.f_fill_styles = &f_fill_styles;
		info.f_sr.f_line_styles = &f_line_styles;
		info.f_sr.f_record = &f_morph_record;
		info.f_save_styles = false;
		ec = SaveShape(info, last_setup);
		if(ec != 0) {
			return ec;
		}
	}
	else {
		if(f_show_bounds || f_show_origin) {
			Vectors		fill_styles;		// empty!
			Vectors		line_styles;		// one style: a red line
			Vectors		record;			// setup + rect.
			Style		red_line;
			Color		red;
			shape_setup_t	bounds_setup(SHAPE_SETUP);
			shape_setup_t	horizontal_setup(SHAPE_SETUP);
			shape_setup_t	vertical_setup(SHAPE_SETUP);
			shape_edges_t	bounds_edges(SHAPE_EDGES);
			shape_edges_t	horizontal_edges(SHAPE_EDGES);
			shape_edges_t	vertical_edges(SHAPE_EDGES);

			// create red line style
			red.Set(255, 0, 0);
			red_line.SetLine(0, 2, red);
			line_styles.Set(-1, &red_line);

			// insert line style
			if(f_show_bounds) {
				bounds_setup.f_fill_ref[0] = 0;
				bounds_setup.f_fill_ref[1] = 0;
				bounds_setup.f_line_ref = 1;
				bounds_setup.f_x = f_bounds[0].XMin();
				bounds_setup.f_y = f_bounds[0].YMin();
				record.Set(-1, &bounds_setup);

				// draw rectangle
				bounds_edges.f_edges.Set(f_bounds[0].XMax() - f_bounds[0].XMin(), 0);
				bounds_edges.f_edges.Set(0, f_bounds[0].YMax() - f_bounds[0].YMin());
				bounds_edges.f_edges.Set(f_bounds[0].XMin() - f_bounds[0].XMax(), 0);
				bounds_edges.f_edges.Set(0, f_bounds[0].YMin() - f_bounds[0].YMax());

				// insert rect.
				record.Set(-1, &bounds_edges);
			}

			if(f_show_origin) {
				// insert X axis
				horizontal_setup.f_fill_ref[0] = 0;
				horizontal_setup.f_fill_ref[1] = 0;
				horizontal_setup.f_line_ref = 1;
				horizontal_setup.f_x = f_bounds[0].XMin();
				horizontal_setup.f_y = 0;
				record.Set(-1, &horizontal_setup);
				horizontal_edges.f_edges.Set(f_bounds[0].XMax() - f_bounds[0].XMin(), 0);
				record.Set(-1, &horizontal_edges);

				// insert Y axis
				vertical_setup.f_fill_ref[0] = 0;
				vertical_setup.f_fill_ref[1] = 0;
				vertical_setup.f_line_ref = 1;
				vertical_setup.f_x = 0;
				vertical_setup.f_y = f_bounds[0].YMin();
				record.Set(-1, &vertical_setup);
				vertical_edges.f_edges.Set(0, f_bounds[0].YMax() - f_bounds[0].YMin());
				record.Set(-1, &vertical_edges);
			}

			// insert all of that in the shape now
			info.f_sr.f_fill_styles = &fill_styles;
			info.f_sr.f_line_styles = &line_styles;
			info.f_sr.f_record = &record;
			ec = SaveShape(info, last_setup);
			if(ec != 0) {
				return ec;
			}
		}
	}

	// save an end of record now
	info.f_data.WriteBits(0, 6);

	SaveTag(data, tag, info.f_data.ByteSize());
	data.Append(info.f_data);

	return 0;
}



int TagShape::SaveWithoutStyles(Data& data)
{
	save_info_t	info;
	int		ec, idx, max;
	shape_setup_t	last_setup(SHAPE_SETUP, true);

	RecordEdges();
	// don't record the last setup!!! that's totally useless!!!

	assert(!f_morph, "can't save a morphing glyph");

	// assume V1.0 for fonts (that's what this is for!)
	info.f_save_alpha = false;
	info.f_ext_size = false;
	info.f_first = false;		// no style is to be saved... (only style changes?)
	info.f_save_styles = false;

	// this is not automatically defined in this case
	info.f_fill_bits_count = 1;
	info.f_line_bits_count = 1;

	max = f_shapes.Count();
	for(idx = 0; idx < max; idx++) {
		info.f_sr = *dynamic_cast<shape_record_t *>(f_shapes.Get(idx));
		ec = SaveShape(info, last_setup);
		if(ec != 0) {
			return ec;
		}
	}

	info.f_sr.f_fill_styles = &f_fill_styles;
	info.f_sr.f_line_styles = &f_line_styles;
	info.f_sr.f_record = &f_record;
	ec = SaveShape(info, last_setup);
	if(ec != 0) {
		return ec;
	}

	// save an end of record now
	info.f_data.WriteBits(0, 6);

	data.PutByte(0x11);		// TODO: should that be 0x11 or dynamically determined?
	data.Append(info.f_data);

	return 0;
}




int TagShape::SaveShape(save_info_t& info, shape_setup_t& last_setup)
{
	int		ec, idx, max;
	shape_what_t	*rec;
	shape_edges_t	*edges;
	shape_setup_t	*setup;

	idx = 0;

	if(info.f_save_styles) {
		if(!info.f_first) {
			// we need a "special" setup entry with the new fill & line styles
			// NOTE: I'm not sure whether it could be merged with the following
			//	 setup record, it is possible to use idx = 1 for that purpose;
			//	 however it would mean that the number of bits used to
			//	 define the new fill & line styles isn't known before you
			//	 read the new styles which happen AFTER the rest!?!? strange
			//	 (stupid?) idea from Macromedia I think
			info.f_data.WriteBits(0x10, 6);		// setup with new styles byte indicator
		}
		info.f_first = false;
		ec = SaveStyles(info);
		if(ec != 0) {
			return ec;
		}
	}

	max = info.f_sr.f_record->Count();
//fprintf(stderr, "Count = %d\n", max);
	while(idx < max) {
		rec = dynamic_cast<shape_what_t *>(info.f_sr.f_record->Get(idx));
		if(rec->f_what == SHAPE_SETUP) {
			// change setup here
			setup = dynamic_cast<shape_setup_t *>(rec);

//fprintf(stderr, "  setup = %p\n", setup);

			ec = SaveSetup(info, *setup, last_setup);
			if(ec != 0) {
				return ec;
			}
		}
		else {
			// we've got a set of edges
			edges = dynamic_cast<shape_edges_t *>(rec);
			edges->f_edges.Save(info.f_data, last_setup.f_x, last_setup.f_y);
		}
		idx++;
	}

	return 0;
}


int TagShape::SaveSetup(save_info_t& info, const shape_setup_t& setup, shape_setup_t& last)
{
	bool		has_line_style, has_fill_style[2], has_move;
	int		sz, bits;
	long		x, y;

	info.f_data.WriteBits(0, 1);		// a setup entry

	// V2.0+ have a NewStyles flag
#if 0
	if(f_version != 1) {
		info.f_data.WriteBits(0, 1);		// has no new styles
	}
#endif
	info.f_data.WriteBits(0, 1);		// has no new styles

	// is the line style setup changing?
	has_line_style = /*!info.f_save_styles ||*/ setup.f_line_ref != -1 && setup.f_line_ref != last.f_line_ref;
	info.f_data.WriteBits(has_line_style, 1);

	// is one of the fill style changing?
	has_fill_style[0] = setup.f_fill_ref[0] != -1 && setup.f_fill_ref[0] != last.f_fill_ref[0];
	has_fill_style[1] = setup.f_fill_ref[1] != -1 && setup.f_fill_ref[1] != last.f_fill_ref[1];
	info.f_data.WriteBits(has_fill_style[1] * 2 + has_fill_style[0], 2);

	// is it moving?
	x = setup.f_x;
	y = setup.f_y;
	if(x == LONG_MIN) {
		x = last.f_x;
	}
	if(y == LONG_MIN) {
		y = last.f_y;
	}
	has_move = x != last.f_x || y != last.f_y;
	info.f_data.WriteBits(has_move, 1);

#if 0
printf("has line: %d has fill: %d/%d, has move: %d (%ld, %ld) vs (%ld, %ld)\n",
		has_line_style, has_fill_style[0], has_fill_style[1], has_move,
		x, y, last.f_x, last.f_y);
#endif

	// save the move if there is one
	if(has_move) {
		sz = SIBitSize(x);
		bits = SIBitSize(y);
		if(bits > sz) {
			sz = bits;
		}
		info.f_data.WriteBits(sz, 5);
		info.f_data.WriteBits(x, sz);
		info.f_data.WriteBits(y, sz);
		last.f_x = x;
		last.f_y = y;
	}

	// save fills if they changed; the # of bits was defined while
	// saving the styles last
	// TODO: if we have to save new fill/line styles we will need
	//	 the new fill/line bits counts BEFORE to save the
	//	 following... (is that logical?!?)
	if(has_fill_style[0]) {
		info.f_data.WriteBits(setup.f_fill_ref[0], info.f_fill_bits_count);
		last.f_fill_ref[0] = setup.f_fill_ref[0];
	}
	if(has_fill_style[1]) {
		info.f_data.WriteBits(setup.f_fill_ref[1], info.f_fill_bits_count);
		last.f_fill_ref[1] = setup.f_fill_ref[1];
	}

	// save the line style if it changed
	if(has_line_style) {
		info.f_data.WriteBits(setup.f_line_ref, info.f_line_bits_count);
		last.f_line_ref = setup.f_line_ref;

		/* tried that -- doesn't seem to be necessary!
		if(info.f_save_styles) {
		}
		else {
			info.f_data.WriteBits(0, info.f_line_bits_count);
printf("Save line style of 0!\n");
		}
		*/
	}

	// we never have new fill & styles here

	return 0;
}


int TagShape::SaveStylesCount(save_info_t& info, long count)
{
	if(f_version == 1) {
		assert(count < 256, "invalid style count of %ld for a V1.0 flash", count);
		info.f_data.PutByte((char) count);
	}
	else if(count > 254) {
		info.f_data.PutByte((char) 255);
		info.f_data.PutShort((short) count);
	}
	else {
		info.f_data.PutByte((char) count);
	}

	return 0;
}


int TagShape::SaveStyles(save_info_t& info)
{
	int		ec, idx, count;
	Style		*style;

// save the FILL STYLES
	count = info.f_sr.f_fill_styles->Count();
	ec = SaveStylesCount(info, count);
	if(ec != 0) {
		return ec;
	}

	for(idx = 0; idx < count; idx++) {
		style = dynamic_cast<Style *>(info.f_sr.f_fill_styles->Get(idx));
		style->Save(info.f_data, info.f_save_alpha, f_morph);
	}

	info.f_fill_bits_count = UIBitSize(count);

// save the LINE STYLES
	count = info.f_sr.f_line_styles->Count();
	ec = SaveStylesCount(info, count);
	if(ec != 0) {
		return ec;
	}

	for(idx = 0; idx < count; idx++) {
		style = dynamic_cast<Style *>(info.f_sr.f_line_styles->Get(idx));
		style->Save(info.f_data, info.f_save_alpha, f_morph);
	}

	info.f_line_bits_count = UIBitSize(count);

// save the number of bits we will use to save fill & line references from now on
	info.f_data.PutByte((info.f_fill_bits_count << 4) + info.f_line_bits_count);

	return 0;
}


