/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;




/////////////////////////////////////////// Envelop
Envelope::Envelope(unsigned long p, unsigned short l, unsigned short r)
{
	f_position = p;
	f_left     = l;
	f_right    = r;
}


Envelope::Envelope(const Envelope& envelope)
{
	f_position = envelope.f_position;
	f_left     = envelope.f_left;
	f_right    = envelope.f_right;
}


unsigned long Envelope::Position(void) const
{
	return f_position;
}


unsigned short Envelope::Left(void) const
{
	return f_left;
}


unsigned short Envelope::Right(void) const
{
	return f_right;
}


void Envelope::Save(Data& data) const
{
	data.PutLong(f_position);
	data.PutShort(f_left);
	data.PutShort(f_right);
}



/////////////////////////////////////////// SoundInfo
SoundInfo::SoundInfo(void)
{
	f_sound_id = 0;
	f_stop = false;
	f_no_multiple = false;
	f_start_position = 0;
	f_end_position = 0;
	f_loop = 1;		// TODO: is this really the default?
	//f_envelopes = ... -- auto init by constructor
}


void SoundInfo::SetSoundID(sswf_id_t id)
{
	f_sound_id = id;
}


void SoundInfo::StopSound(bool stop)
{
	f_stop = stop;
}


void SoundInfo::NoMultiple(bool no_multiple)
{
	f_no_multiple = no_multiple;
}


void SoundInfo::SetRange(unsigned long start, unsigned long end)
{
	assert(start < end || start == 0 || end == 0, "SoundInfo::SetRange(): expects a larger end than start");
	f_start_position = start;
	f_end_position = end;
}


void SoundInfo::SetLoop(unsigned short loop)
{
	f_loop = loop;
}


void SoundInfo::AddEnvelope(const Envelope& envelope)
{
	Envelope	*e;

	if(f_envelopes.Count() >= 255) {
		// TODO: use our new error object whenever it is created
		fprintf(stderr, "ERROR: too many envelopes for SoundInfo object.\n");
		return;
	}

	e = new Envelope(envelope);
	MemAttach(e, sizeof(Envelope), "SoundInfo::AddEnvelop(): duplication of the envelop object");
	f_envelopes.Set(-1, e);
}


void SoundInfo::Save(Data& data) const
{
	int		idx, max;

	data.PutShort(f_sound_id);
	data.WriteBits(0, 2);
	data.WriteBits(f_stop, 1);
	if(f_stop) {
		// if the sound is being stopped, we don't need any other info
		data.WriteBits(0, 5);
	}
	else {
		max = f_envelopes.Count();

		data.WriteBits(f_no_multiple, 1);
		data.WriteBits(max > 0, 1);
		data.WriteBits(f_loop != 1, 1);
		data.WriteBits(f_start_position != 0, 1);
		data.WriteBits(f_end_position != 0, 1);

		if(f_start_position != 0) {
			data.PutLong(f_start_position);
		}
		if(f_end_position != 0) {
			data.PutLong(f_end_position);
		}
		if(f_loop != 1) {
			data.PutShort(f_loop);
		}
		// when we have envelops, save them all!
		if(max > 0) {
			data.PutByte(max);
			for(idx = 0; idx < max; idx++) {
				dynamic_cast<const Envelope *>(f_envelopes.Get(idx))->Save(data);
			}
		}
	}
}


/////////////////////////////////////////// TagStartSound
TagStartSound::TagStartSound(TagBase *parent)
	: TagBase("start_sound", parent)
{
	f_sound_info = 0;
}



TagBase::swf_type_t TagStartSound::TypeFlags(void) const
{
	return SWF_TYPE_CONTROL | SWF_TYPE_REFERENCE;
}


void TagStartSound::SetInfo(SoundInfo *sound_info)
{
	// NOTE: should we copy the user object?
	f_sound_info = sound_info;
}


int TagStartSound::PreSave(void)
{
	if(f_sound_info == 0) {
		return 1;
	}

	return 0;
}



int TagStartSound::Save(Data& data)
{
	Data		sub_data;

	// the sound info size varies with envelopes
	f_sound_info->Save(sub_data);

	SaveTag(data, SWF_TAG_START_SOUND, sub_data.ByteSize());
	data.Append(sub_data);

	return  0;
}








/////////////////////////////////////////// TagSound
#define	SNDINFMT(width, endian)		((width)|((endian)&7))

const int TagSound::g_sound_rates[4] = { 5512, 11025, 22050, 44100 };
const int TagSound::g_bitrates[2][16] =
{
	{ -1, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, -1 },
	{ -1, 8,  16, 24, 32, 40, 48, 56,  64,  80,  96, 112, 128, 144, 160, -1 }
};
const int TagSound::g_frequencies[4][4] =
{
	{ 11025, -1, -1, -1 },
	{    -1, -1, -1, -1 },
	{ 22050, -1, -1, -1 },
	{ 44100, -1, -1, -1 }
};


TagSound::TagSound(TagBase *parent)
	: TagBaseID("sound", parent)
{
	// the output compression format
	f_format = SOUND_FORMAT_RAW;

	// the following are all set at the time the SetData() function
	// is called -- it is done that way so the coherency of the
	// data is preserved (changing from stereo to mono would require
	// changing the data!)
	f_stereo = false;
	f_rate = 0;
	f_width = 0;
	f_samples = 0;
	f_data_size = 0;
	f_data_maxsize = 0;
	f_data = 0;
	f_latency_seek = 0;
}


TagBase::swf_type_t TagSound::TypeFlags(void) const
{
	// TODO: with streaming sound effects, the object may have references
	//	 though it will be to itself...
	return SWF_TYPE_DEFINE | SWF_TYPE_HAS_ID;
}




int TagSound::SetFilename(const char *filename)
{
	int	ec;
	FILE	*f;

	f = fopen(filename, "rb");
	if(f == NULL) {
		return -1;
	}

	ec = LoadWaveFile(f);
	if(ec != 0) {
		rewind(f);
		ec = LoadMP3File(f);
	}

	fclose(f);

	return ec;
}

int TagSound::LoadWaveFile(FILE *f)
{
	unsigned char	header[12];
	sound_wave_t	fmt;
	unsigned long	sz, data_size;
	char		*data;
	int		cnt;

	if(fread(header, 12, 1, f) != 1) {
		return -1;
	}

	if(header[0] != 'R' || header[1] != 'I' || header[ 2] != 'F' || header [ 3] != 'F'
	|| header[8] != 'W' || header[9] != 'A' || header[10] != 'V' || header [11] != 'E') {
		return -1;
	}

	// size is in header[4..7] -- we don't really need it do we?

	// we go through the list of chunks until we got the fmt
	// and data chunks
	data = 0;
	data_size = 0;	// if data is defined, this is too -- avoid warnings
	cnt = 0;
	while(cnt != 3) {
		// this read will fail if we reach the end of the file
		// before we've got the fmt and data chunks
		if(fread(header, 8, 1, f) != 1) {
			MemFree(data);
			return -1;
		}
		if(header[0] == 'f' && header[1] == 'm' && header[2] == 't' && header[3] == ' ') {
			/* we found the format info, read it! */
			if((cnt & 1) != 0) {
				// Ooooops! more than one fmt chunk?!?
				MemFree(data);
				return -1;
			}
			sz = header[4] + (header[5] << 8)
					+ (header[6] << 16) + (header[7] << 24);
			if(sz != sizeof(fmt)) {
				// incompatible/unknown format chunk size
				MemFree(data);
				return -1;
			}
			if(fread(&fmt, sizeof(fmt), 1, f) != 1) {
				// can't read format info
				MemFree(data);
				return -1;
			}
			// on big endian systems it is necessary to swap values here
#if BYTE_ORDER == BIG_ENDIAN
			fmt.format   = swap_short(fmt.format);
			fmt.channels = swap_short(fmt.channels);
			fmt.rate     = swap_int(fmt.rate);
			fmt.width    = swap_short(fmt.width);
			// what isn't swapped we don't use
			//fmt.align  = swap_short(fmt.align);
			//fmt.average_rate = swap_int(fmt.average_rate);
#endif
			// quickly check to see if we understand this info
			if(fmt.format != 1	// only PCM files understood at this time!
			|| (fmt.channels != 1 && fmt.channels != 2)
			|| (fmt.width != 8 && fmt.width != 16)) {
				MemFree(data);
				return -1;
			}
			cnt |= 1;
		}
		else if(header[0] == 'd' && header[1] == 'a' && header[2] == 't' && header[3] == 'a') {
			if((cnt & 2) != 0) {
				// Ooooops! more than one data chunk?!?
				MemFree(data);
				return -1;
			}
			data_size = header[4] + (header[5] << 8)
					+ (header[6] << 16) + (header[7] << 24);
			data = (char *) MemAlloc(data_size, "TagSound::SetFilename(): temporary data buffer");
			if(fread(data, data_size, 1, f) != 1) {
				// can't read data samples
				MemFree(data);
				return -1;
			}
			cnt |= 2;
		}
	}

	// got the format info and (uncompressed) data now
	// note that 8 bits data is unsigned whereas 16 bits
	// is signed!!! data is always in little endian which
	// thus need to be swapped on big endian computers
	return SetData(
		data,
		data_size,
		SOUND_ENDIAN_LITTLE,
		fmt.width == 8 ? 8 : -16,
		fmt.rate,
		fmt.channels == 2);
}



int TagSound::LoadMP3File(FILE *f)
{
	unsigned char	header[4];
	int		ec, frame_size;

	// TODO: we certainly should check whether
	//	 this looks like an MP3 file...

	// That's an MP3 (hmmm)
	f_format = SOUND_FORMAT_MP3;

	// Not too sure about this one at this time we kind of ignore
	// this parameter...
	//f_latency_seek = 0;

	// MP3s are always 16 bits
	f_width = 16;

	// Restart from the start each time a new file is loaded
	f_data_size = 0;
	f_samples = 0;

	// save the total size of the buffer at this time
	//for(int cnt = 0; cnt  < 50; cnt++) {
	for(;;) {
		// get the size of frame
		ec = CheckMP3Header(f, header, frame_size);
		if(ec != 0) {
			break;
		}

		// now that we have the size, make sure the data
		// buffer is large enough
		if(f_data_size + frame_size > f_data_maxsize) {
			// large blocks will speed things up for large MP3 files
			f_data_maxsize = (f_data_size + frame_size + 1024 * 1024 - 1) & -1024 * 1024;
			f_data = (unsigned char *) MemRealloc(f_data, f_data_maxsize, "TagSound::LoadMP3File() -- frame buffer");
		}

		// copy the header
		f_data[f_data_size + 0] = header[0];
		f_data[f_data_size + 1] = header[1];
		f_data[f_data_size + 2] = header[2];
		f_data[f_data_size + 3] = header[3];

		// finally read the frame in our buffer
		if(fread(f_data + f_data_size + 4, 1, frame_size - 4, f) != (size_t) (frame_size - 4)) {
			return -1;
		}

		// move to the next frame
		f_data_size += frame_size;
	}

	// the value 2 represents EOF
	return ec == 2 ? 0 : ec;
}


int TagSound::CheckMP3Header(FILE *f, unsigned char *header, int& frame_size)
{
	int	ec, mode;
	int	bitrate, frequency;

	// Read the first or next header
	// (we always try to re-sync.)
	ec = ReadMP3Header(f, header);
	if(ec != 0) {
		return ec;
	}

        // Only layer 3 files are valid for the flash player
        if(((header[1] >> 1) & 3) != 1) {
		return -1;
	}

	mode = (header[1] >> 3) & 3;

	// Get bitrate from bitrates array -- depends on mode as well
	bitrate = g_bitrates[mode == 3 ? 0 : 1][(header[2] >> 4) & 15];

	// Get frequency -- also depends on mode
	frequency = g_frequencies[mode][(header[2] >> 2) & 3];

	// make sure these are valid
	if(bitrate == -1 || frequency == -1) {
		return -1;
	}

	// WARNING: not too sure whether the frequence and stereo
	// information should reflect the first or last MP3 frame
	// at this time it's the last (but I would guess it won't
	// change very much in a single MP3?)
	switch(frequency) {
	//case  5512: f_rate = 0; break; -- this is not supported in MP3
	case 11025: f_rate = 1; break;
	case 22050: f_rate = 2; break;
	case 44100: f_rate = 3; break;

	// we can't change the MP3 rate without decompressing
	// and recompressing which we don't currently support
	// [also the table only gives us the three valid values
	// anyway...]
	default:
		return -1;

	}

	// Channel mode
	f_stereo = ((header[3] >> 6) & 3) != 3;

	// Increment the samples counter
	f_samples += mode == 3 ? 1152 : 576; // V2 and V1 sample counts

	// Calculate the frame size (including the pad)
	frame_size = (mode == 3 ? 144 : 72) * bitrate * 1000 / frequency
			+ ((header[2] >> 1) & 1);

	return 0;
}


// Note: if we can't find the next Sync. then
//	 we tell the caller that we found the
//	 end of the file
int TagSound::ReadMP3Header(FILE *f, unsigned char *header)
{
	int	r;

	// TODO: we may want to load the entire file at once
	//	 so it is much faster than this -- imagine that
	//	 we read a 2Mb file 1 byte at a time...

	// sync to first or next frame
	do {
		r = fread(header, 1, 1, f);
		if(r != 1) {
			goto err;
		}
	} while(header[0] != 255);

	do {
		r = fread(header + 1, 1, 1, f);
		if(r != 1) {
			goto err;
		}
	} while((header[1] & 0xE0) != 0xE0);

	r = fread(header + 2, 1, 2, f);
	if(r == 2) {
		return 0;
	}
err:
	return r >= 0 ? 2 : -1;
}


void TagSound::SetFormat(sound_format_t format)
{
	// SetFormat() doesn't work when it's already MP3...
	// [this is because we don't support the compression
	// we only support loading an existing MP3 and saving
	// it as is...]
	if(f_format == SOUND_FORMAT_MP3) {
		return;
	}

	f_format = format;
}


short TagSound::ReadSample(const unsigned char *data, unsigned short adjust, int in_fmt)
{
	switch(in_fmt) {
	case SNDINFMT(8, SOUND_ENDIAN_LITTLE):
	case SNDINFMT(8, SOUND_ENDIAN_BIG):
	case SNDINFMT(8, SOUND_ENDIAN_SAME):
	case SNDINFMT(8, SOUND_ENDIAN_DONTUSE):
		// 8 bits are returned on 8 bits.
		return (signed char) (data[0] + adjust);

	case SNDINFMT(16, SOUND_ENDIAN_LITTLE):
#if BYTE_ORDER == LITTLE_ENDIAN
	case SNDINFMT(16, SOUND_ENDIAN_SAME):
#endif
		return (data[0] + data[1] * 256) + adjust;

	case SNDINFMT(16, SOUND_ENDIAN_BIG):
#if BYTE_ORDER == BIG_ENDIAN
	case SNDINFMT(16, SOUND_ENDIAN_SAME):
#endif
		return (data[1] + data[0] * 256) + adjust;

	case SNDINFMT(24, SOUND_ENDIAN_LITTLE):
#if BYTE_ORDER == LITTLE_ENDIAN
	case SNDINFMT(24, SOUND_ENDIAN_SAME):
#endif
		return (data[1] + data[2] * 256) + adjust;

	case SNDINFMT(24, SOUND_ENDIAN_BIG):
#if BYTE_ORDER == BIG_ENDIAN
	case SNDINFMT(24, SOUND_ENDIAN_SAME):
#endif
		return (data[2] + data[1] * 256) + adjust;

	case SNDINFMT(32, SOUND_ENDIAN_LITTLE):
#if BYTE_ORDER == LITTLE_ENDIAN
	case SNDINFMT(32, SOUND_ENDIAN_SAME):
#endif
		return (data[2] + data[3] * 256) + adjust;

	case SNDINFMT(32, SOUND_ENDIAN_BIG):
#if BYTE_ORDER == BIG_ENDIAN
	case SNDINFMT(32, SOUND_ENDIAN_SAME):
#endif
		return (data[1] + data[0] * 256) + adjust;

	}

	/* the list of valid formats is checked sooner so we should
	 * never reach this place!
	 */

	/*NOTREACHED*/
	return 0;	/* avoid warnings */
}



void TagSound::Resample(unsigned char *snd, unsigned int out_bytes, const unsigned char *src, size_t size, unsigned int in_bytes, size_t max, double fix, unsigned short adjust, int in_fmt)
{
	size_t			idx;
	double			p1, p2, start, end, s, inv_fix;
	unsigned long		pi1, pi2, cnt;
	short			sample;
	const unsigned char	*in, *end_src;

	out_bytes -= f_width == 16 ? 2 : 1;

	inv_fix = 1.0 / fix;
	end_src = src + size;
	for(idx = 0; idx < max; idx++) {
		// start position & end position
		p1 = (double) idx * inv_fix;
		p2 = p1 + inv_fix;

		// check with integral values
		pi1 = (unsigned long) floor(p1);
		pi2 = (unsigned long) floor(p2);

		if(pi1 == pi2) {
			// in this case we only use part of one sample
			start = p2 - p1;
			end = 0.0;
			cnt = 0;
		}
		else {
			// in this case we need multiple samples
			// we may also use part of two samples
			// (when cnt == 0)
			start = 1.0 - p1 + (double) pi1;
			end = p2 - (double) pi2;
			cnt = pi2 - pi1 - 1;
		}

		// skip to the 1st byte we want to read
		in = src + pi1 * in_bytes;

		// read the first sample (note that start can be 1.0)
		assert(in < end_src, "TagSound::Resample(): source data pointer out of bounds");
		s = (double) ReadSample(in, adjust, in_fmt) * start;
		in += in_bytes;

		// when multiple samples are required to compute the
		// new sample, then read them all right here
		while(cnt > 0) {
			cnt--;
			assert(in < end_src, "TagSound::Resample(): source data pointer out of bounds");
			s += (double) ReadSample(in, adjust, in_fmt);
			in += in_bytes;
		}

		// if we have something at the end (end can be 0.0 but never 1.0)
		// and we still have samples in the input, use that bit too
		if(end > 0.0 && in < end_src) {
			s += (double) ReadSample(in, adjust, in_fmt) * end;
		}

		// note that there can't be an overflow
		// or I'm really bad at math! so the
		// following works without having to test
		// for overflows
		sample = (short) rint(s * fix);

		*snd++ = (unsigned char) sample;
		if(f_width == 16) {
			*snd++ = (unsigned char) (sample >> 8);
		}

		// in stereo the following skips the other channel too
		snd += out_bytes;
	}
}


size_t TagSound::GetSamplesCount(void) const
{
	return f_samples;
}


int TagSound::SetData(const void *data, size_t size, sound_endian_t endian, int width, unsigned int rate, bool stereo)
{
	unsigned int		in_bytes, out_bytes, good_rate, in_fmt;
	unsigned short		sample, adjust;
	unsigned char		*snd;
	size_t			in_samples, in_count, cnt, out_samples;
	double			fix;
	bool			signed_data;

	// SetData() doesn't work yet for MP3
	if(f_format == SOUND_FORMAT_MP3) {
		return -1;
	}

	if((signed_data = width < 0)) {
		width = -width;
	}

	// in release mode, these are checked below and the
	// function will return 1.
	assert(width == 8 || width == 16 || width == 24 || width == 32, "TagSound::SetData(): the width of a sound must be 8, 16, 24 or 32, the value %d can't be used", width);
	assert(rate >= 1024 && rate <= 88000, "TagSound::SetData(): the samples rate must be between 1024 and 88000, the value %d is being rejected", rate);
	assert(endian >= 0 && endian < SOUND_ENDIAN_UNKNOWN, "TagSound::SetData(): unknown endianess: %d", endian);

	/* first we compute what we need (i.e. the size of the buffer we will keep) */
	switch(width) {		// bytes per samples
	case  8: in_bytes = 1; out_bytes = 1; f_width =  8; break;
	case 16: in_bytes = 2; out_bytes = 2; f_width = 16; break;
	case 24:
	case 32: in_bytes = 4; out_bytes = 2; f_width = 16; break;
	default: return -1;
	}

	/* compute the number of samples in the input buffer */
	in_count = in_samples = size / in_bytes;

	if((f_stereo = stereo)) {
		  in_bytes *= 2;
		 out_bytes *= 2;
		in_samples /= 2;
	}

	// assert the # of samples, not the size since in_samples could be
	// zero after the division of size by in_bytes
	assert(data != 0 && in_samples > 0, "TagSound::SetData(): a sound data buffer can't be empty or a null pointer");
	if(data == 0 || in_samples == 0) {
		return -1;
	}

	/* compute a correct rate if not already so */
	switch(rate) {
	case  5512: f_rate = 0; break;
	case 11025: f_rate = 1; break;
	case 22050: f_rate = 2; break;
	case 44100: f_rate = 3; break;
	default:
		if(rate < 1024) {
			return -1;
		}
		if(rate < 6063) {
			f_rate = 0;
		}
		else if(rate < 12127) {
			f_rate = 1;
		}
		else if(rate < 24255) {
			f_rate = 2;
		}
		else if(rate > 88000) {
			return -1;
		}
		else {
			f_rate = 3;
		}
	}
	good_rate = g_sound_rates[f_rate];

	/* if the rate is the same, then we can compute the output buffer as is */
	if(good_rate == rate) {
		/* use input buffer as is (beside endianess) */
		out_samples = in_samples;
		fix = 0.0;		// unused in this case, avoid warnings
	}
	else {
		/* oops! user had a strange input file? got to fix the samples */
		fix = (double) good_rate / (double) rate;
		out_samples = (size_t) ceil((double) in_samples * fix);
	}

	in_fmt = SNDINFMT(width, endian);

	switch(in_fmt) {
	case SNDINFMT(8, SOUND_ENDIAN_LITTLE):
	case SNDINFMT(8, SOUND_ENDIAN_BIG):
	case SNDINFMT(8, SOUND_ENDIAN_SAME):
	case SNDINFMT(8, SOUND_ENDIAN_DONTUSE):
		adjust = signed_data ? 0 : -128;
		break;

	case SNDINFMT(16, SOUND_ENDIAN_LITTLE):
	case SNDINFMT(16, SOUND_ENDIAN_BIG):
	case SNDINFMT(16, SOUND_ENDIAN_SAME):
	case SNDINFMT(24, SOUND_ENDIAN_LITTLE):
	case SNDINFMT(24, SOUND_ENDIAN_BIG):
	case SNDINFMT(24, SOUND_ENDIAN_SAME):
	case SNDINFMT(32, SOUND_ENDIAN_LITTLE):
	case SNDINFMT(32, SOUND_ENDIAN_BIG):
	case SNDINFMT(32, SOUND_ENDIAN_SAME):
		adjust = signed_data ? 0 : -32768;
		break;

	default:
		return -1;

	}

	/* the f_data buffer size can now be computed */
	MemFree(f_data);
	f_samples = 0;
	f_data = (unsigned char *) MemAlloc(out_samples * out_bytes, "TagSound::SetData() -- array of samples");

	/* if we have the same rate, do it quick! */
	snd = f_data;
	if(good_rate == rate) {
		cnt = in_count;
		if(stereo) {
			// we do all at once, so we need to have the proper
			// number of bytes not taking the stereo in account
			in_bytes /= 2;
		}
		do {
			// read one sample
			sample = ReadSample((const unsigned char *) data, adjust, in_fmt);
			data = (char *) data + in_bytes;
			// save in our buffer in little endian and increase
			// the snd pointer accordingly
			*snd++ = (unsigned char) sample;
			if(f_width == 16) {
				*snd++ = (unsigned char) (sample >> 8);
			}
			cnt--;
		} while(cnt > 0);
	}
	else if(stereo) {
		// resample stereo sound (resample left then right track)
		Resample(snd, out_bytes, (const unsigned char *) data, size, in_bytes, out_samples, fix, adjust, in_fmt);
		Resample(snd + out_bytes / 2, out_bytes, (const unsigned char *) data + in_bytes / 2, size, in_bytes, out_samples, fix, adjust, in_fmt);
	}
	else {
		// resample non-stereo sound (one call to the loop is enough)
		Resample(snd, out_bytes, (const unsigned char *) data, size, in_bytes, out_samples, fix, adjust, in_fmt);
	}

	f_samples = out_samples;

	return 0;
}



void TagSound::SetMono(void)
{
	unsigned char	*in, *out;
	size_t		idx;
	unsigned short	sample;

	// SetMone() doesn't work yet for MP3
	if(f_format == SOUND_FORMAT_MP3) {
		return;
	}

	if(f_stereo && f_samples > 0) {
		// the user is asking us to remove the L+R and generate a Mono
		// track instead, so what we do is Mono = (L+R)/2
		in = out = f_data;
		if(f_width == 8) {
			for(idx = 0; idx < f_samples; idx++, in += 2, out++) {
				out[0] = (in[0] + in[1]) / 2;
			}
		}
		else {
			for(idx = 0; idx < f_samples; idx++, in += 4, out += 2) {
				sample = (in[0] + in[2] + ((in[1] + in[3]) << 8)) / 2;
				out[0] = (unsigned char) sample;
				out[1] = (unsigned char) (sample >> 8);
			}
		}
		f_stereo = false;
	}
}



void TagSound::Set8Bits(void)
{
	unsigned char	*in, *out;
	size_t		idx;

	// Set8Bits() doesn't work yet for MP3
	if(f_format == SOUND_FORMAT_MP3) {
		return;
	}

	if(f_width == 16 && f_samples > 0) {
		// the user is asking us to make the data 8 bits instead of 16
		// this is really simple: we only keep the top 8 bits!
		in = out = f_data;
		if(f_stereo) {
			for(idx = 0; idx < f_samples; idx++, in += 4, out += 2) {
				out[0] = in[1];
				out[1] = in[3];
			}
		}
		else {
			for(idx = 0; idx < f_samples; idx++, in += 2, out++) {
				out[0] = in[1];
			}
		}
		f_width = 8;
	}
}


int TagSound::PreSave(void)
{
	// no effect with an undefined sound effect
	if(f_samples <= 0) {
		return 0;
	}

	if(f_width == 8 && f_format != SOUND_FORMAT_RAW && f_format != SOUND_FORMAT_UNCOMPRESSED) {
		// any compressed data must be 16 bits,
		// we need a function to transform 8 bits to 16 bits
		// (which may mean a decompression + re-compression)
		return -1;
	}

	switch(f_format) {
	case SOUND_FORMAT_RAW:
	case SOUND_FORMAT_ADPCM:
		MinimumVersion(2);
		break;

	case SOUND_FORMAT_MP3:
	case SOUND_FORMAT_UNCOMPRESSED:
		MinimumVersion(4);
		break;

	case SOUND_FORMAT_NELLYMOSER:
		MinimumVersion(6);
		// TODO: until this is supported, we return an error
		return -1;
		break;

	default:
		return -1;

	}

	return 0;
}


int TagSound::Save(Data& data)
{
	Data		sub_data;
	unsigned int	size;

	// no effect with an undefined sound effect
	if(f_samples <= 0) {
		return 0;
	}

	// TODO: the following doesn't support streaming sound...

	SaveID(sub_data);

	// info
	sub_data.WriteBits(f_format, 4);
	sub_data.WriteBits(f_rate, 2);
	sub_data.WriteBits(f_width == 16 ? 1 : 0, 1);
	sub_data.WriteBits(f_stereo, 1);

	switch(f_format) {
	case SOUND_FORMAT_RAW:			// though we don't check the processor, we save this in little endian!!!
	case SOUND_FORMAT_UNCOMPRESSED:		// always little endian (V4.x)
		// the size in bytes of the f_data buffer
		size = f_samples;
		if(f_stereo) {
			size *= 2;
		}
		if(f_width == 16) {
			size *= 2;
		}
		sub_data.PutLong(f_samples);
		sub_data.Write(f_data, size);
		break;

	case SOUND_FORMAT_MP3:
		sub_data.PutLong(f_samples);
		sub_data.PutShort(f_latency_seek);
		sub_data.Write(f_data, f_data_size);
		break;

	// TODO: add support for ADPCM & NellyMoser

	default:
		assert(0, "the f_format (%d) is accepted in PreSave() but not in Save()", f_format);
		/*NOTREACHED*/

	}

	SaveTag(data, SWF_TAG_DEFINE_SOUND, sub_data.ByteSize());
	data.Append(sub_data);

	return 0;
}




